# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes that wrap objects returned by Discord.

A **dataclass** is a class that contains data and allows the user to edit classes. These classes are
stateful,  as in they store state which allows the user to edit them as appropriate.

.. currentmodule:: curious.dataclasses

.. autosummary::
    :toctree: dataclasses

    appinfo
    bases
    channel
    embed
    emoji
    guild
    invite
    member
    message
    permissions
    presence
    reaction
    role
    search
    user
    voice_state
    webhook
    widget

"""
